"""Re-usable plot components.

This module contains several re-usable plot components that contain the data for a
Plotly (sub) figure. See :obj:`~ragraph.plot.generic.Component` for the definition of a
plot component.
"""

from ragraph.plot.components.blank import Blank  # noqa
from ragraph.plot.components.labels import Labels  # noqa
from ragraph.plot.components.legend import Legend  # noqa
from ragraph.plot.components.piemap import PieMap  # noqa
from ragraph.plot.components.tree import Tree  # noqa



"""Ford climate control system.

This dataset describes a climate control system as it was to be found in Ford vehicles.
Four different dependency types have been documented, being spatial, energy flow,
information flow, and material flow dependencies. These are weighted from -2 to 2, where
the following definitions have been used:

    * +2: Required
    * +1: Desired
    *  0: Indifferent
    * -1: Undesired
    * -2: Detrimental

We have added an "adjacency" weight, which is the nonnegative sum of all dependencies
between components.

Reference: Pimmler, T. U., & Eppinger, S. D. (1994). Integration Analysis of Product
Decompositions. ASME Design Theory and Methodology Conference.
"""


edge_weights = [
    "adjacency",
    "spatial",
    "energy flow",
    "information flow",
    "material flow",
]



"""
##########
Heuristics
##########

Heuristics are combinations of algorithms available in the other subpackages or have an
output that does not strictly fit one of the other categories. Not all heuristics have a
common argument structure because of their more diverse nature.

********************
Available heuristics
********************

The following heuristics are directly available after importing
:obj:`ragraph.analysis.heuristics`:

* :func:`johnson() <ragraph.analysis.heuristics.johnson.circuit_finding_algorithm>`:
  Johnson's circuit finding algorithm. Generates all cycles in a graph, including
  overlapping and nested cycles.
* :func:`markov_gamma()
  <ragraph.analysis.heuristics.markov_gamma.markov_gamma_clustering>`: Combination of
  :obj:`Gamma bus detection <ragraph.analysis.bus.gamma.gamma_bus_detection>` and
  :obj:`(Hierarchical) Markov Clustering <ragraph.analysis.cluster.markov>`. Detects a
  tree structure with buses on global or local levels.
"""
# flake8: noqa, ignore errors on unused imports here.

from ragraph.analysis.heuristics._johnson import circuit_finding_algorithm as johnson
from ragraph.analysis.heuristics._markov_gamma import (
    markov_gamma_clustering as markov_gamma,
)



"""Ford climate control system clustered using the Markov-Gamma clustering and bus
detection heuristic.

Reference: Pimmler, T. U., & Eppinger, S. D. (1994). Integration Analysis of Product
Decompositions. ASME Design Theory and Methodology Conference.
"""


edge_weights = [
    "adjacency",
    "spatial",
    "energy flow",
    "information flow",
    "material flow",
]



"""
#####################
Sequencing algorithms
#####################

Sequencing algorithms search for sequences of nodes according to some objective or
logic. Simple are sorting by name and more elaborate examples sort according to Markov
transisition probabilities or sequences without feedback (edges to earlier nodes).

There are several metrics from literature available to grade sequences as well. The
metrics are documented over at :obj:`metrics`.

Finally, there are some utilities like branch-sorting (recursively sorting all branches
in a hierarchical tree instead of all leaf nodes as a whole) available in :obj:`utils`.

********************
Available algorithms
********************

The following algorithms are directly accessible after importing
:obj:`ragraph.analysis.sequence`:

* :func:`markov() <ragraph.analysis.sequence.markov.markov_sequencing>`: Markov
  sequencing based on relative node dependency and influence.
* :func:`name() <ragraph.analysis.sequence.name.name>`: Sequence nodes by node name.
* :func:`tarjans_dfs() <ragraph.analysis.sequence.tarjan.tarjans_dfs>`: Sequence nodes
  according to Tarjan's Depth First Search algorithm. Sequence a Directed Acyclic Graphs
  (DAG) into a sequence without feedback edges to earlier nodes.
"""
# flake8: noqa, ignore errors on unused imports here.

from ragraph.analysis.sequence import metrics, utils
from ragraph.analysis.sequence._axis import axis_sequencing as axis
from ragraph.analysis.sequence._markov import markov_sequencing as markov
from ragraph.analysis.sequence._name import sort_by_name as name
from ragraph.analysis.sequence._scc_tearing import scc_tearing
from ragraph.analysis.sequence._tarjan import tarjans_dfs_sequencing as tarjans_dfs
from ragraph.analysis.sequence.utils import branchsort



"""Ford Motor Company Hood Development Process.

This DSM application was made to test the effectiveness of a DSM analysis to improve a
highly developed iterative design process in terms of:

- Reduction of product development lead time
- Reduction of product development lead-time variation


As it is a process DSM, the nodes represent tasks in the design process. The edges
represent dependencies between them. Each task in the design process was assigned a task
volatility value, which is an indication of the probability of rework.


Node (task) weights:

- EC(i) and EC(r) represent the initial and rework costs
- ED(i) and ED(r) represent the initial and rework durations
- Information variability (likelihood of input changing)


Edge weights:

- volatility: an indication of the probability of rework when the dependent task is
  executed without properly waiting for its input to finish. It's value was determined
  using the information variability of the source (depended on) task and the sensitivity
  of the target (dependent) task.


References:
Browning, Tyson R., and Steven D. Eppinger. 2002, November. Modeling Impacts of Process
Architecture on Cost and Schedule Risk in Product Development. IEEE Transactions on
Engineering Management 49 (4):428-442.
Krishnan, Viswanathan, Steven D. Eppinger, and Daniel E. Whitney. 1997, April.
A Model-Based Framework to Overlap Product Development Activities.
Management Science 43 (4):437-451.
Yassine, Ali A., Daniel E. Whitney, and Tony P. Zambito. 2001. Assessment of Rework
Probabilities for Simulating Product Development Processes Using the Design Structure
Matrix (DSM). ASME Design Engineering Technical Conferences, DTM-21693.
"""


node_weights = [
    "EC(i) ($000)",
    "EC(r) ($000)",
    "ED(i) (days)",
    "ED(r) (% of ED (i))",
    "information variability",
]
edge_weights = ["volatility"]



"""Example graph for similarity analysis. Products are related (have edges) to
attributes that are of importance in product portfolio analysis. Edges have edge kind
"incidence". Nodes have kind "product" or "attribute".
"""



"""Pratt & Whitney Aircraft Engine

A directed graph describing the Pratt & Whitney PW4098 commercial high bypass-ratio
turbofan engine. The graph describes a combination of the actual hardware dependencies
of the engine and those of the development teams involved with them. It is a weighted
and directed graph featuring 60 elements, four of which are sometimes left out as they
represent the integration teams and no individual hardware components. Weak and strong
dependencies are distinguished using weights of 1 and 2, respectively.

Reference:
Rowles, C. M. (1999). System integration analysis of a large commercial aircraft engine.
"""

edge_weights = ["adjacency"]



"""Elevator system decomposed into 175 components.

The Complex Elevator System is described using an undirected graph with multiple
dependency types. It describes a machine-room-less elevator called the 'Kone MonoSpace'.
It is designed for low- to midrise buildings and uses permanent-magnet electric motors.
The five defined edge types are spatial, material, mechanical energy, electrical energy
and information. It was published in a variation of 175 elements and a less granular
variation of 45 elements. The less granular variation collapses a set of pre-defined
modules and therefore contains less detail of the system.

Reference:
Niutanen, V., Hölttä-otto, K., Rahardjo, A., & Stowe, H. M. (2017). Complex Elevator
System DSM - Case for a DSM Design Sprint. In 19th International dependency and
structure modeling conference, DSM 2017.
"""

edge_weights = [
    "spatial",
    "material",
    "information",
    "mechanical energy",
    "electrical energy",
    "sum",
    "binary",
]



"""Example 8 node directed graph from Shaja and Sudhakar.

Reference:
Shaja, A. S., & Sudhakar, K. (2010). Optimized sequencing of analysis components in
multidisciplinary systems. Research in Engineering Design, 21(3), 173–187.
https://doi.org/10.1007/s00163-009-0082-5
"""

edge_weights = ["adjacency"]



"""Elephant Specification Language format support."""

try:
    from raesl.compile import to_graph as from_esl  # noqa
except ImportError:
    raise ImportError(
        "Package 'raesl' needs to be available to convert the ESL format."
    )



from typing import Any, Callable, List, Tuple

import numpy as np
from ragraph.analysis import cluster, similarity
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


class SimilarityAnalysis:
    """Similarity analysis of nodes based upon mutual mapping relations.

    Arguments:
        cols: List of column nodes.
        rows: List of row nodes.
        edges: List of edges from column nodes to row nodes to be used in similarity
            analysis.
        col_sim_threshold: Column similarity threshold. Values below this threshold are
            pruned from the similarity matrix and the corresponding edges are removed.
            Defaults to 0.0 (no threshold).
        row_sim_threshold: Column similarity threshold. Values below this threshold are
            pruned from the similarity matrix and the corresponding edges are removed.
            Defaults to 0.0 (no threshold).

    Class Attributes:
        similarity_kind: Edge kind for similarity edges. Defaults to ``similarity``.

    Note:
        A mapping matrix relating M column nodes to N row nodes is used as input for
        the similarity analysis.
    """

    similarity_kind = "similarity"

    def __init__(
        self,
        rows: List[Node],
        cols: List[Node],
        edges: List[Edge],
        col_sim_threshold: float = 0.0,
        row_sim_threshold: float = 0.0,
    ):
        self._cols: List[Node] = []
        self._rows: List[Node] = []
        self._edges: List[Edge] = []
        self._col_sim_threshold = 0.0
        self._row_sim_threshold = 0.0
        self._graph: Graph = None  # type: ignore

        self.cols = cols
        self.rows = rows
        self.edges = edges
        self.col_sim_threshold = col_sim_threshold
        self.row_sim_threshold = row_sim_threshold

    @property
    def rows(self) -> List[Node]:
        """List of rows nodes."""
        return self._rows

    @rows.setter
    def rows(self, value: List[Node]):
        kinds = set([n.kind for n in value])
        if len(kinds) > 1:
            raise ValueError(
                f"""All row nodes must be of the same kind. Found node kinds:
                 {kinds}."""
            )

        if self._rows:
            self._rows = value
            self.update_graph()
        else:
            self._rows = value

    @property
    def cols(self) -> List[Node]:
        """List of column nodes."""
        return self._cols

    @cols.setter
    def cols(self, value: List[Node]):
        kinds = set([n.kind for n in value])
        if len(kinds) > 1:
            raise ValueError(
                f"""All columns nodes must be of the same kind. Found node kinds:
                 {kinds}."""
            )

        if self._cols:
            self._cols = value
            self.update_graph()
        else:
            self._cols = value

    @property
    def edges(self) -> List[Edge]:
        """List of edges."""
        return self._edges

    @edges.setter
    def edges(self, value: List[Edge]):
        if self._edges:
            self._edges = value
            self.update_graph()
        else:
            self._edges = value

    @property
    def row_sim_threshold(self) -> float:
        """Similarity threshold. Values below this threshold are pruned from the
        row similarity matrix and the corresponding edges are removed.
        """
        return self._row_sim_threshold

    @row_sim_threshold.setter
    def row_sim_threshold(self, value: float):
        if value < 0.0 or value >= 1.0:
            raise ValueError(
                f"""Similarity threshold must have a value between 0.0 and 1.0. Found
                value: {value}."""
            )

        if self._row_sim_threshold is not None:
            self._row_sim_threshold = float(value)
            self.update_row_similarity()
        else:
            self._row_sim_threshold = float(value)

    @property
    def col_sim_threshold(self) -> float:
        """Similarity threshold. Values below this threshold are pruned from the
        column similarity matrix and the corresponding edges are removed.
        """
        return self._col_sim_threshold

    @col_sim_threshold.setter
    def col_sim_threshold(self, value: float):
        if value < 0.0 or value >= 1.0:
            raise ValueError(
                f"""Similarity threshold must have a value between 0.0 and 1.0. Found
                value: {value}."""
            )

        if self._col_sim_threshold is not None:
            self._col_sim_threshold = float(value)
            self.update_col_similarity()
        else:
            self._col_sim_threshold = float(value)

    @property
    def graph(self) -> Graph:
        """Graph containing similarity edges."""
        if self._graph:
            return self._graph
        else:
            self.update_graph()
            return self._graph

    @property
    def row_similarity_matrix(self) -> np.ndarray:
        """The row similarity matrix based on their mapping row."""
        jaccard = similarity.jaccard_matrix(self.rows, self.row_mapping)
        jaccard[jaccard < self.row_sim_threshold] = 0.0
        return jaccard

    @property
    def col_similarity_matrix(self) -> np.ndarray:
        """The column similarity matrix based on their mapping column."""
        jaccard = similarity.jaccard_matrix(self.cols, self.col_mapping)
        jaccard[jaccard < self.col_sim_threshold] = 0.0
        return jaccard

    def update_graph(self) -> None:
        """Update Internal similarity graph"""
        self._graph = Graph(nodes=self.cols + self.rows, edges=self.edges)
        self.update_col_similarity()
        self.update_row_similarity()

    def update_row_similarity(self) -> None:
        """Update Jaccard Row Similarity Index edges between (clustered) rows."""
        self._update_similarity(self.rows, self.row_similarity_matrix)

    def update_col_similarity(self) -> None:
        """Update Jaccard Column Similarity Index edges between (clustered) columns."""
        self._update_similarity(self.cols, self.col_similarity_matrix)

    def _update_similarity(self, nodes: List[Node], mat: np.ndarray) -> None:
        """Update Jaccard Similarity Index edges between (clustered) nodes."""
        if not self.graph:
            self.update_graph()

        for e in [
            edge
            for edge in self.graph.edges_between_all(nodes, nodes)
            if edge.kind == self.similarity_kind
        ]:
            self.graph.del_edge(e)

        for row, target in enumerate(nodes):
            for col, source in enumerate(nodes):
                if row == col:
                    continue
                sim = mat[row][col]
                if not sim:
                    continue
                self.graph.add_edge(
                    Edge(
                        source,
                        target,
                        kind=self.similarity_kind,
                        weights=dict(similarity=mat[row][col]),
                    )
                )

    def cluster_rows(
        self,
        algo: Callable[[Graph, Any], Tuple[List[Node]]] = cluster.markov,
        **algo_args,
    ) -> None:
        """Cluster column nodes based on their similarity. Updates Graph in-place.

        Arguments:
            algo: Clustering algorithm. Should take a graph as first argument and
                cluster it in-place. Defaults to :obj:`cluster.markov`.
            **algo_args: Algorithm arguments. See :obj:`cluster.markov` for sensible
                defaults.
        """
        self._cluster(self.rows, algo, **algo_args)

    def cluster_cols(
        self,
        algo: Callable[[Graph, Any], Tuple[List[Node]]] = cluster.markov,
        **algo_args,
    ) -> None:
        """Cluster column nodes based on their similarity. Updates Graph in-place.

        Arguments:
            algo: Clustering algorithm. Should take a graph as first argument and
                cluster it in-place. Defaults to :obj:`cluster.markov`.
            **algo_args: Algorithm arguments. See :obj:`cluster.markov` for sensible
                defaults.
        """
        self._cluster(self.cols, algo, **algo_args)

    def _cluster(
        self,
        leafs: List[Node],
        algo: Callable[[Graph, Any], Tuple[List[Node]]] = cluster.markov,
        **algo_args,
    ) -> None:
        """Cluster column nodes based on their similarity. Updates Graph in-place.

        Arguments:
            leafs: List of row or column nodes to be clustered.
            algo: Clustering algorithm. Should take a graph as first argument and
                cluster it in-place. Defaults to :obj:`cluster.markov`.
            **algo_args: Algorithm arguments. See :obj:`cluster.markov` for sensible
                defaults.
        """
        algo(self.graph, leafs=leafs, **algo_args, inplace=True)  # type: ignore

    def row_mapping(self, row: Node) -> List[bool]:
        """Boolean possession checklist for a row node w.r.t. :obj:`self.cols`."""
        return [self.check_mapping(col, row) for col in self.cols]

    def col_mapping(self, col: Node) -> List[bool]:
        """Boolean possession checklist for a column node w.r.t. :obj:`self.rows`."""
        return [self.check_mapping(col, row) for row in self.rows]

    def check_mapping(self, col: Node, row: Node) -> bool:
        """Check whether a column node maps to a row node."""
        if len(self.graph.directed_edges[col.name][row.name]) > 0:
            return True
        else:
            return False



"""
##########
Comparison
##########

Comparison provides classes for comparing :obj:`Graph <ragraph.graph.Graph` objects
to find the commonalities (sigma) and differences (delta).
"""
# flake8: noqa, ignore errors on unused imports here.

from ragraph.analysis.comparison._delta import DeltaAnalysis
from ragraph.analysis.comparison._sigma import SigmaAnalysis
from ragraph.analysis.comparison.utils import (
    check_edge_equality,
    get_common_leafs,
    get_unique_edges,
    get_unique_leafs,
)



from typing import List

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


def get_unique_leafs(a: Graph, b: Graph) -> List[Node]:
    """Get unique leafs in graph a.

    Arguments:
        a: Graph object.
        b: Graph object.

    Returns:
        List of unique leafs in Graph a.
    """
    return [
        a.node_dict[n]
        for n in a.node_dict.keys() - b.node_dict.keys()
        if a.node_dict[n].is_leaf
    ]


def get_common_leafs(a: Graph, b: Graph, ue: List[Edge]) -> List[Node]:
    """Get common leafs in graph a and b

    Arguments:
      a: Graph object.
      b: Graph object.
      ue: List of unique edges.

    Returns:
      List of common leafs in graphs a and b.
    """
    # Shared leafs in both graphs, which do NOT participate in unique edges.
    common_leafs = [
        a.node_dict[n]
        for n in (a.node_dict.keys() & b.node_dict.keys())
        if a.node_dict[n].is_leaf
    ]
    return common_leafs


def check_edge_equality(a: Edge, b: Edge) -> bool:
    """Checks if two edges are equal based on source, target, kind, labels.

    Arguments:
        a: Edge a.
        b: Edge b

    Returns:
        Whether edges are considered equal.
    """
    return (
        a.source.name == b.source.name
        and a.target.name == b.target.name
        and a.kind == b.kind
        and set(a.labels) == set(b.labels)
    )


def get_unique_edges(a: Graph, b: Graph) -> List[Edge]:
    """Get unique edges in graph a.

    Arguments:
        a: Graph object.
        b: Graph object.

    Returns:
        List of unique edges in a.
    """
    unq = [
        ea for ea in a.edges if not any(check_edge_equality(ea, eb) for eb in b.edges)
    ]
    return unq



"""Minimal example to illustrate the compatibility analysis.

Contains 6 component variant nodes. They are divided in three node kinds (e.g.
components), which correspond to the first character in their node names: A1, B1, B2,
C1, C2, C3. For ease of usage, the "performance" weight of each node is set to it's node
name's second character.

Compatibility between nodes is signalled using edges with a "compatibility" kind.
"""

node_weights = ["performance"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ragraph',
 'ragraph.analysis',
 'ragraph.analysis.bus',
 'ragraph.analysis.cluster',
 'ragraph.analysis.comparison',
 'ragraph.analysis.compatibility',
 'ragraph.analysis.heuristics',
 'ragraph.analysis.sequence',
 'ragraph.analysis.similarity',
 'ragraph.colors',
 'ragraph.datasets',
 'ragraph.datasets.aircraft_engine',
 'ragraph.datasets.architecture_integral',
 'ragraph.datasets.architecture_mix',
 'ragraph.datasets.architecture_modular',
 'ragraph.datasets.climate_control',
 'ragraph.datasets.climate_control_mg',
 'ragraph.datasets.compatibility',
 'ragraph.datasets.design',
 'ragraph.datasets.eefde_lock',
 'ragraph.datasets.elevator175',
 'ragraph.datasets.elevator45',
 'ragraph.datasets.esl',
 'ragraph.datasets.esl.pump',
 'ragraph.datasets.ford_hood',
 'ragraph.datasets.kodak3d',
 'ragraph.datasets.ledsip',
 'ragraph.datasets.localbus',
 'ragraph.datasets.overlap',
 'ragraph.datasets.pathfinder',
 'ragraph.datasets.shaja8',
 'ragraph.datasets.similarity',
 'ragraph.datasets.tarjans8',
 'ragraph.datasets.ucav',
 'ragraph.io',
 'ragraph.io.archimate',
 'ragraph.io.xml',
 'ragraph.plot',
 'ragraph.plot.components']

package_data = \
{'': ['*']}

install_requires = \
['dd>=0.5.7,<0.6.0',
 'lxml>=4.8.0,<5.0.0',
 'numpy>=1.22.3,<2.0.0',
 'plotly>=5.7.0,<6.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0']

extras_require = \
{'esl': ['raesl'], 'plot': ['kaleido==0.2.1']}

setup_kwargs = {
    'name': 'ragraph',
    'version': '1.16.0',
    'description': 'Ratio graph handling in Python.',
    'long_description': "#######\nRaGraph\n#######\n\nRaGraph is a package to create, manipulate, and analyze graphs consisting of nodes and\nedges. Nodes usually represent (hierarchies of) objects and edges the dependencies or\nrelationships between them.\n\nThese graphs, or networks if you will, lend themselves well to applied analyses like\nclustering and sequencing, as well as analyses involving the calculation of various\ninsightful metrics.\n\n\n**********\nQuickstart\n**********\n\nInstallation\n============\n\nRaGraph can be installed using ``pip install ragraph`` for any Python version >=3.9. Or,\nfor Poetry managed projects, use ``poetry add ragraph`` to add it as a dependency.\n\n\nUsing RaGraph\n=============\n\nRaGraph's primary use is working with Graph objects that contain Nodes and Eges between\nNodes. See the `usage documentation <https://ragraph.ratio-case.nl/usage/index.html>`_\nfor more info!\n\n\n***************\nDeveloper guide\n***************\n\nPython packaging information\n============================\n\nThis project is packaged using `poetry <https://python-poetry.org/>`_. Packaging\ninformation as well as dependencies are stored in `pyproject.toml <./pyproject.toml>`_.\n\nInstalling the project and its development dependencies can be done using ``poetry install``.\n\n\nInvoke tasks\n============\n\nMost elemental maintenance tasks can be accomplished using\n[Invoke](https://www.pyinvoke.org/). After installing using ``poetry install`` and\nenabling the environment using ``poetry shell``, you can run all tasks using ``inv\n[taskname]`` or ``invoke [taskname]``. E.g. ``inv docs`` builds the documentation.\n\n\nVersioning\n==========\n\nThis project uses `semantic versioning <https://semver.org>`_. Version increments are\nchecked using `Raver <https://raver.ratio-case.nl>`_.\n\n\nChangelog\n=========\n\nChangelog format as described by https://keepachangelog.com/ has been adopted.\n",
    'author': 'Ratio Innovations B.V.',
    'author_email': 'info@ratio-case.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/ratio-case/python/ragraph',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
